<?php
namespace Opencart\Catalog\Controller\Extension\Vtbpay\Payment;

require __DIR__ . '/vendor/autoload.php';

use \Vtbpay\Classes\Api\VtbApi,
    \Vtbpay\Classes\Common\VtbPayLogger,
    \Vtbpay\Classes\Exception\VtbPayException,
    \Opencart\System\Engine\Registry,
    \Opencart\System\Engine\Controller,
    \Opencart\System\Library\Log;

class Vtbpay extends Controller
{
    private VtbPayLogger $logger;
    private array $order;


    /**
     * VtbPay constructor
     *
     * Set up the logging functionality for the VtbPay transactions.
     *
     * @param object $registry The registry of the OpenCart system
     */
    public function __construct(Registry $registry)
    {
        parent::__construct($registry);
        $this->setVtbpayLogger();
        $this->load->language('extension/vtbpay/payment/vtbpay');
    }


    /**
     * Get VTB API instance
     *
     * Returns an instance of the VTB API using configuration values.
     *
     * @return VtbApi An instance of the VtbApi
     */
    private function getVtbApi(): VtbApi
    {
        return new VtbApi(
            $this->config->get('payment_vtbpay_client_id'),
            $this->config->get('payment_vtbpay_client_secret'),
            (bool) $this->config->get('payment_vtbpay_test_mode'),
            $this->config->get('payment_vtbpay_merchant_authorization') ?: ''
        );
    }


    /**
     * Index function
     *
     * Prepares data for the checkout view, and returns the rendered HTML.
     *
     * @return string Rendered HTML of the checkout view
     */
    public function index(): string
    {
        $data['button_confirm'] = $this->language->get('button_confirm');
        $data['action'] = $this->url->link('extension/vtbpay/payment/vtbpay.confirm', '', true);
        $data['logo'] = '/extension/vtbpay/catalog/view/image/payment/vtbpay.png';
        $data['css'] = '/extension/vtbpay/catalog/view/stylesheet/payment/vtbpay.css';
        $data['js'] = '/extension/vtbpay/catalog/view/js/payment/vtbpay.js';
        $data['desc'] = $this->config->get('payment_vtbpay_desc');
        return $this->load->view('extension/vtbpay/payment/vtbpay', $data);
    }


    /**
     * Confirm function
     *
     * Initiates a payment transaction with VTB Pay and redirects the user to the VTB Pay payment page.
     * Logs all VTB Pay transactions.
     *
     * @return void
     */
    public function confirm(): void
    {
        $order_id = $this->session->data['order_id'] ?? null;
        $this->logger->setOption('additionalCommonText', 'payment-' . rand(1111, 9999));

        try {
            if (!$order_id) {
                throw new \Exception($this->language->get('error_no_order_id'));
            }

            $this->setOrder($order_id);

            if (empty($pay_url = $this->getPayUrl())) {
                throw new \Exception($this->language->get('error_pay_url_not_available'));
            }

            unset($this->session->data['order_id']);

            $this->response->redirect($pay_url);

        } catch (\Exception | VtbPayException $e) {
            // Handle exception and log error
            $this->executeErrorScenario(
                $e,
                $order_id,
                __FUNCTION__
            );
        }
    }


    /**
     * Получает URL-адрес платежа для перенаправления.
     *
     * @return string URL-адрес платежа.
     */
    private function getPayUrl(): string
    {
        $order_id = $this->order['order_id'];
        $email = $this->order['email'] ?: $this->config->get('payment_vtbpay_email_fiscal');
        $total = $this->order['total'];
        $return_url = $this->url->link('extension/vtbpay/payment/vtbpay.return'); // Get return URL
        $two_stage = (bool) $this->config->get('payment_vtbpay_two_stage');
        $enable_fiscal = (bool) $this->config->get('payment_vtbpay_enable_fiscal');

        $items = [];
        if (!$two_stage && $enable_fiscal) $items = $this->getItems();

        $this->model_checkout_order->addHistory(
            $order_id,
            1,
            'orderId: ' . $order_id,
            true
        );

        // Logging input
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderLink - INPUT: ', [
            'arguments' => [
                'order_id' => $order_id,
                'email' => $email,
                'time' => time(),
                'total' => $total,
                'return_url' => $return_url,
                'two_stage' => $two_stage,
                'items' => $items
            ]
        ]);

        $response = $this->getVtbApi()->getOrderLink(
            $order_id,
            $email,
            time(),
            $total,
            $return_url,
            $two_stage,
            $items
        );

        // Logging output
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderLink - OUTPUT: ', [
            'response' => $response
        ]);

        return $response['object']['payUrl'] ?? '';
    }


    /**
     * Получает массив товаров заказа и доставку.
     *
     * @return array Массив товаров.
     */
    private function getItems(): array
    {
        $this->load->model('catalog/product');

        // Настройки по умолчанию для фискальных параметров
        $fiscal_settings = [
            'payment_vtbpay_measure_fiscal' => 0,
            'payment_vtbpay_tax_type_fiscal' => 'none',
            'payment_vtbpay_payment_type_fiscal' => 'full_prepayment',
            'payment_vtbpay_payment_subject_fiscal' => 1
        ];
        $items = [];
        $key = 1;
        // Добавление в массив товаров
        foreach ($this->order['products'] as $item) {
            $product = $this->model_catalog_product->getProduct($item['product_id']);

            $product_fiscal = [];
            // Получение фискальных параметров для каждого продукта
            foreach ($fiscal_settings as $name => $default) {
                $product_fiscal[$name] = $product[$name] ?: '-';
                if ($product_fiscal[$name] === '-') $product_fiscal[$name] = $this->config->get($name) ?: $default;
            }

            $items[] = [
                'positionId' => $key,
                'name' => $item['name'],
                'code' => (string) $item['product_id'],
                'price' => self::parse_amount($item['price']),
                'measure' => (int) $product_fiscal['payment_vtbpay_measure_fiscal'],
                'quantity' => (int) $item['quantity'],
                'taxParams' => [
                    'taxType' => $product_fiscal['payment_vtbpay_tax_type_fiscal']
                ],
                'paymentType' => $product_fiscal['payment_vtbpay_payment_type_fiscal'],
                'paymentSubject' => (int) $product_fiscal['payment_vtbpay_payment_subject_fiscal'],
                'amount' => self::parse_amount($item['total'])
            ];

            $key++;
        }

        // Добавление в массив доставки
        foreach ($this->order['totals'] as $total) {
            if ($total['code'] == 'shipping' && $total['value'] > 0) {

                $shipping_amount = self::parse_amount($total['value']);
                $items[] = [
                    'positionId' => $key,
                    'name' => $total['title'],
                    'price' => $shipping_amount,
                    'measure' => 0,
                    'quantity' => 1,
                    'taxParams' => [
                        'taxType' => $this->config->get('payment_vtbpay_tax_type_fiscal') ?: 'none'
                    ],
                    'paymentType' => $this->config->get('payment_vtbpay_payment_type_delivery_fiscal') ?: 'full_prepayment',
                    'paymentSubject' => 4,
                    'amount' => $shipping_amount
                ];

                $key++;
            }
        }

        return $items;
    }


    /**
     * Преобразует строку в число с плавающей запятой.
     *
     * @param mixed $amount Сумма для преобразования.
     * @return float Преобразованное значение суммы.
     */
    private static function parse_amount($amount): float
    {
        if (is_string($amount)) {
            $amount = str_replace([' ', ','], ['', '.'], $amount);
        }

        return floatval($amount);
    }


    /**
     * Return function
     *
     * Processes the return from VTB Pay, updates the order status based on the return data,
     * and prepares and displays a success or failure message to the user.
     *
     * @return void
     */
    public function return(): void
    {
        $order_id = $this->request->get['orderId'] ?? null;
        $this->logger->setOption('additionalCommonText', 'return-' . rand(1111, 9999));

        // Logging $_REQUEST
        $this->logger->debug(
            'redirect > ' . __FUNCTION__ . ': ', [
            'request_data' => $_REQUEST
        ]);

        try {
            if (!$order_id) {
                throw new \Exception($this->language->get('error_no_order_id'));
            }

            $this->setOrder($order_id);

            // Check if the order belongs to the current authorized user
            if ($this->order['customer_id'] != $this->customer->getId()) {
                throw new \Exception($this->language->get('error_unauthorized_order_access'));
            }

            $this->changeStatus();

            $this->cart->clear();
            $this->response->redirect($this->url->link('checkout/success'));

        } catch (\Exception | VtbPayException $e) {
            // Handle exception and log error
            $this->executeErrorScenario(
                $e,
                $order_id,
                __FUNCTION__
            );
        }
    }


    /**
     * Обработчик callback, вызываемый платёжкой при изменение статуса платежа.
     *
     * @return void
     */
    public function webhook(): void
    {
        $php_input = json_decode(file_get_contents('php://input'), true) ?: null;
        $this->logger->setOption('additionalCommonText', 'webhook-' . rand(1111, 9999));

        // Logging php input
        $this->logger->debug(
            'callback > ' . __FUNCTION__ . ': ', [
            'php_input' => $php_input
        ]);

        $order_id = $php_input['object']['orderId'] ?? '';

        try {
            if (empty($order_id)) {
                throw new \Exception($this->language->get('error_no_order_id'));
            }

            $this->setOrder($order_id);

            $this->changeStatus();

            $this->response->setOutput('OK');

        } catch (\Exception | VtbPayException $e) {
            // Handle exception and log error
            $this->executeErrorScenario(
                $e,
                $order_id,
                __FUNCTION__,
                false
            );
        }
    }


    /**
     * Получени информации о платеже и зменение статуса заказа
     *
     * @return void
     */
    public function changeStatus(): void
    {
        $payment_status = $this->getPaymentStatus();
        $current_order_status = $this->order['order_status_id'];
        $order_id = $this->order['order_id'];
        $failed_order_status = $this->config->get('payment_vtbpay_failed_transaction_status');
        $availableStatuses = [
            'PAID' => $this->config->get('payment_vtbpay_successful_transaction_status'),
            'PENDING' => '1',
            'REFUNDED' => '13'
        ];
        $order_status = $availableStatuses[$payment_status] ?? '';

        if (!empty($order_status)) {
            if ($current_order_status !== $order_status) {
                $this->model_checkout_order->addHistory(
                    $order_id,
                    $order_status,
                    $payment_status,
                    true
                );
            }
        }
        else {
            if ($current_order_status !== $failed_order_status) {
                $this->model_checkout_order->addHistory(
                    $order_id,
                    $failed_order_status,
                    $payment_status
                );
            }
            throw new \Exception($this->language->get('error_payment_not_paid'));
        }
    }


    /**
     * Получение статуса заказа из VTB API.
     *
     * @return string Значение статуса заказа, если оно существует, или пустая строка в противном случае.
     */
    private function getPaymentStatus(): string
    {
        $order_id = $this->order['order_id'];

        // Logging input
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderInfo - INPUT: ', [
            'arguments' => [
                'order_id' => $order_id
            ]
        ]);
        $response = $this->getVtbApi()->getOrderInfo($order_id);

        // Logging output
        $this->logger->debug(
            __FUNCTION__ . ' > getOrderInfo - OUTPUT: ', [
            'response' => $response
        ]);

        return $response['object']['status']['value'] ?? '';
    }


    /**
     * Инициализируем свойство order результатом выполнения функции getOrder.
     *
     * @param int $order_id Идентификатор заказа.
     *
     * @return self|Exception Объект заказа, если он существует, или Exception в противном случае.
     */
    private function setOrder(int $order_id): self
    {
        $this->load->model('checkout/order');

        $this->order = $this->model_checkout_order->getOrder($order_id);
        if (!$this->order) {
            throw new \Exception($this->language->get('error_order_not_found'));
        }

        $this->order['products'] = $this->model_checkout_order->getProducts($order_id);
        $this->order['totals'] = $this->model_checkout_order->getTotals($order_id);

        return $this;
    }


    /**
     * Сценарий, выполняемый после отлова исключения
     *
     * @param \Exception|VtbPayException $e   Объект исключения.
     * @param int $order_id                   Идентификатор заказа.
     * @param string $caller_func             Название функции, котоорая вызывает обработчик исключений.
     * @param bool $make_redirect             Совершать редирект на страницу ошибки или нет.
     *
     * @return void
     */
    private function executeErrorScenario(
        $e,
        $order_id,
        $caller_func = '',
        $make_redirect = true
    ): void {
        $context = [];
        if (method_exists($e, 'getContext')) $context = $e->getContext();

        $this->logger->error(sprintf(
            $caller_func . ' > ' . __FUNCTION__ . ' > VtbPay exception : %s; Order id: %s;',
            $e->getMessage(),
            $order_id ?: ''
        ), $context);

        if ($make_redirect) {
            $this->session->data['error'] = $e->getMessage();
            $this->response->redirect($this->url->link('checkout/failure', '', true));
        }
        else {
            $this->response->setOutput($e->getMessage());
        }
    }


    /**
     * Инициализация и настройка объекта класса VtbPayLogger.
     *
     * Эта функция инициализирует и настраивает логгер, используемый плагином VtbPay для ведения журнала.
     *
     * @return void
     */
    private function setVtbpayLogger(): void
    {
        $oc_log = new Log('vtbpay-' . date('d-m-Y') . '.log');

        // Настройка в админке (логироать или нет)
        $logging = $this->config->get('payment_vtbpay_logging');

        $this->logger = VtbPayLogger::getInstance()
                        ->setOption('showCurrentDate', false)
                        ->setCustomRecording(function($message) use ($oc_log){
                            $oc_log->write($message);
                        }, VtbPayLogger::LOG_LEVEL_ERROR)
                        ->setCustomRecording(function($message) use ($oc_log, $logging){
                            if ($logging) $oc_log->write($message);
                        }, VtbPayLogger::LOG_LEVEL_DEBUG);
    }
}
